/*
** This file is a part of DSSL library.
**
** Copyright (C) 2008, SSLTech LLC
** Copyright (C) 2005-2008, Vladimir Shcherbakov <vladimir@ssltech.net>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
*/
#include "stdinc.h"
#include "stream.h"
#include "session.h"

#ifdef _DEBUG
const char* GetStreamStateName( TcpStreamState state )
{
	switch( state )
	{
		case tcpClosed:			return "CLOSED";
		case tcpListen:			return "LISTEN";
		case tcpSynSent:		return "SYN-SENT";
		case tcpSynReceived:	return "SYN-RECEIVED";
		case tcpEstablished:	return "ESTABLISHED";
		case tcpFinWait1:		return "FIN-WAIT-1";
		case tcpFinWait2:		return "FIN-WAIT-2";
		case tcpCloseWait:		return "CLOSE-WAIT";
		case tcpClosing:		return "CLOSING";
		case tcpLastACK:		return "LAST-ACK";
		case tcpTimeWait:		return "TIME_WAIT";
	}

	return "UNKNOWN";
}
#endif

/* Local prototypes */
static void ClearStateChangeData( TcpStream* stream );


void StreamInit( TcpStream* stream, TcpSession* sess, uint32_t ip, uint16_t port, TcpStreamState initState )
{
	_ASSERT( stream );

	stream->ip_addr = ip;
	stream->port = port;
	stream->pktHead = NULL;
	stream->pktTail = NULL;
	stream->state = initState;
	stream->nextSeqExpected = 0;
	stream->session = sess;
	stream->lastACK = 0;
	ClearStateChangeData( stream );
}

static int StreamGetPacketCount( TcpStream* stream )
{
	int cnt = 0;
	DSSL_Pkt* pkt = stream->pktHead;

	while( pkt )
	{
		cnt++;
		pkt = pkt->next;
	}

	return cnt;
}


void StreamFreeData( TcpStream* stream )
{
	DSSL_Pkt* pkt = stream->pktHead;

#ifdef NM_TRACE_TCP_STREAMS
	DEBUG_TRACE2( "\nFreeStreamData: stream %p; %d packets freed",
		stream, StreamGetPacketCount( stream ) );
#endif

	while( pkt ) 
	{
		DSSL_Pkt* t = pkt->next;
		PktFree( pkt );
		pkt = t;
	}

	stream->pktTail = stream->pktHead = NULL;
	stream->nextSeqExpected = 0;

}


static void StreamInsertAfter( TcpStream* stream, DSSL_Pkt* pktInsert, DSSL_Pkt* pktAfter )
{
#ifdef NM_TRACE_TCP_STREAMS
	{
		uint32_t seq = PKT_TCP_SEQ( pktInsert );
		uint32_t seq_after = PKT_TCP_SEQ( pktAfter );
		DEBUG_TRACE2( "\nInsert seq:%u  after: %u", seq, seq_after );
	}
#endif

	pktInsert->prev = pktAfter;
	pktInsert->next = pktAfter->next;
	pktAfter->next = pktInsert;

	if( pktInsert->next ) { pktInsert->next->prev = pktInsert; }
	if( pktAfter == stream->pktTail ) { stream->pktTail = pktInsert; }
}


static void StreamInsertBefore( TcpStream* stream, DSSL_Pkt* pktInsert, DSSL_Pkt* pktBefore )
{
	_ASSERT( pktBefore );

#ifdef NM_TRACE_TCP_STREAMS
	{
		uint32_t seq = PKT_TCP_SEQ( pktInsert );
		uint32_t seq_before = PKT_TCP_SEQ( pktBefore );
		DEBUG_TRACE2( "\nInsert seq:%u  before: %u", seq, seq_before );
	}
#endif

	pktInsert->prev = pktBefore->prev;

	if( pktBefore->prev )
	{
		_ASSERT( pktBefore->prev->next == pktBefore );
		pktBefore->prev->next = pktInsert;
	}
	else
	{
		_ASSERT( pktBefore == stream->pktHead );
		stream->pktHead = pktInsert;
	}

	pktBefore->prev = pktInsert;
	pktInsert->next = pktBefore;
}


static int IsNextPacket( TcpStream* stream, DSSL_Pkt* pkt )
{
	_ASSERT( stream );
	_ASSERT( pkt );

	return stream->nextSeqExpected == PKT_TCP_SEQ( pkt )
		&& stream->lastACK >= PktNextTcpSeqExpected(pkt);
}


static int StreamConsumePacket( TcpStream* stream, DSSL_Pkt* pkt )
{	
#ifdef NM_TRACE_TCP_STREAMS
		DEBUG_TRACE2( "\nConsuming seq:%u, len: %d", PKT_TCP_SEQ( pkt ), pkt->data_len );
#endif
	stream->nextSeqExpected = PktNextTcpSeqExpected( pkt );
	stream->session->packet_time = pkt->pcap_header.ts;
	return stream->session->OnNewPacket( stream, pkt );
}

static int StreamEnqueue( TcpStream* stream, DSSL_Pkt* pkt )
{
	DSSL_Pkt* p = NULL;
	uint32_t seq = PKT_TCP_SEQ( pkt );
	int processed = 0;

	if( seq < stream->nextSeqExpected )
	{
#ifdef NM_TRACE_TCP_STREAMS
		DEBUG_TRACE2( "\nDropping a packet (retransmission), seq:%u, next_seq: %u",
			seq, stream->nextSeqExpected );
#endif
		return DSSL_RC_OK;
	}

	/* simple case - first packet in the list */
	if( stream->pktHead == NULL )
	{
#ifdef NM_TRACE_TCP_STREAMS
		DEBUG_TRACE1( "\nAdding the head at seq:%u", seq );
#endif
		_ASSERT( stream->pktTail == NULL);

		p = PktClone( pkt );
		p->next = p->prev = NULL;
		stream->pktHead = stream->pktTail = p;
		processed = 1;
	}
	else
	{
		p = stream->pktHead;
		/* find where this packet fits */
		while( p && !processed )
		{
			uint32_t seq_p = PKT_TCP_SEQ( p );
			if( seq_p == seq )
			{
	#ifdef NM_TRACE_TCP_STREAMS
				DEBUG_TRACE2( "\nDropping retransmission, seq:%u, len: %d", seq, (int) pkt->data_len );
	#endif
				processed = 1;
			}
			else if( seq_p > seq )
			{
				/* Insert in front*/
				StreamInsertBefore( stream, PktClone( pkt ), p );
				processed = 1;
			}
			else
			{
				p = p->next;
			}
		}

		/* add to the tail */
		if( !processed )
		{
			StreamInsertAfter( stream, PktClone( pkt ), stream->pktTail );
			processed = 1;
		}
	}

	_ASSERT( processed );
	return DSSL_RC_OK;
}


static DSSL_Pkt* StreamDequeue( TcpStream* stream )
{
	DSSL_Pkt* retval = stream->pktHead;
	
	if( stream->pktHead ) { stream->pktHead = stream->pktHead->next; }
	if( stream->pktHead ) {
		stream->pktHead->prev = NULL;
	} else {
		stream->pktTail = NULL;
	}

	return retval;
}

/*
========================================================
	StreamProcessPacket: Main packet processing routine 
========================================================
*/

int StreamProcessPacket( TcpStream* stream, DSSL_Pkt* pkt )
{
	int rc = DSSL_RC_OK;

	if( pkt->tcp_header->th_flags & TH_SYN )
	{
		_ASSERT( pkt->data_len == 0 );
		stream->nextSeqExpected = PktNextTcpSeqExpected( pkt );
		return rc;
	}

	/* ignore any 0-length packets other TH_SYN */
	if( pkt->data_len == 0 ) { return rc; }

	/* is this the next packet in sequence? */
	if( IsNextPacket( stream, pkt ) )
	{
		rc = StreamConsumePacket( stream, pkt );
	}
	else
	{
		rc = StreamEnqueue( stream, pkt );
	}

	/* check any previoulsy enqueued out-of-order packets are ready to be processed */
	while( rc == DSSL_RC_OK && stream->pktHead && IsNextPacket( stream, stream->pktHead ) )
	{
		DSSL_Pkt* p = StreamDequeue( stream );
		rc = StreamConsumePacket( stream, p );
		/* 
			packets are cloned when they are put into the queue, 
			so they should be freed here 
		*/
		PktFree(p); 
	}

	return rc;
}


static void ClearStateChangeData( TcpStream* stream )
{
	stream->stch_action = stchNoAction;
	stream->stch_flags = 0;
	stream->stch_ack = 0;
	stream->stch_state = tcpClosed;
}


void StreamStateChangeOnACK( TcpStream* stream, TcpStreamState state, uint32_t ack )
{
	stream->stch_action = stchCheckACK;
	stream->stch_ack = ack;
	stream->stch_state = state;
}


void StreamStateChangeOnFlags( TcpStream* stream, TcpStreamState state, uint8_t flags )
{
	stream->stch_action = stchCheckFlags;
	stream->stch_flags = flags;
	stream->stch_state = state;
}


static void STREAM_CHANGE_STATE( TcpStream* stream, TcpStreamState state )
{
	SET_STREAM_STATE( stream, state );
	ClearStateChangeData( stream );
}


void StreamCheckStateChange( TcpStream* stream, DSSL_Pkt* pkt )
{
	switch( stream->stch_action )
	{
	case stchCheckACK:
		if( ntohl( pkt->tcp_header->th_ack) >= stream->stch_ack )
		{
			STREAM_CHANGE_STATE( stream, stream->stch_state );
		}
		break;

	case stchCheckFlags:
		if( pkt->tcp_header->th_flags & stream->stch_flags )
		{
			STREAM_CHANGE_STATE( stream, stream->stch_state );
		}
		break;


	default:
		_ASSERT( FALSE );
	case stchNoAction:
		break;
	}
}

int StreamAcceptACK(TcpStream* stream, uint32_t ack)
{
	int rc = DSSL_RC_OK;
	stream->lastACK = ack;

	/* check any previoulsy enqueued out-of-order packets are ready to be processed */
	while( rc == DSSL_RC_OK && stream->pktHead && IsNextPacket( stream, stream->pktHead ) )
	{
		DSSL_Pkt* p = StreamDequeue( stream );
		rc = StreamConsumePacket( stream, p );
		/* 
			packets are cloned when they are put into the queue, 
			so they should be freed here 
		*/
		PktFree(p); 
	}

	return rc;
}
